/**
   CSE791 Programming Assignment #1
   Simple Image Smoothing

   smoother.c --- implement a five-point stencil program

**/

#include <stdlib.h>
#include <mpi.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <pnm.h>

#define WORKTAG 1


/**
   function asssembling()

**/ 

void assembling(xel** _xel, int numOfSide, int colP, int rowP, int formatP, xelval maxvalP, char* filename) {

  int i,j,k,y,z;
  int offsetX, offsetY;
  int sizeXBlock, sizeYBlock;
  int srcProcess;
  char cp[30], savefile[35];
  char *tmpfilename, *extention;

  FILE *fp;

  xel *oneDimXel;
  MPI_Status status;

  offsetX = 0;
  offsetY = 0;

  for(i=0;i<numOfSide;i++) {
    for(j=0;j<numOfSide;j++) {
      
      // calculate source process
      srcProcess = i*numOfSide + j;

      // recv X,Y size and rendered image
      MPI_Recv(&sizeXBlock,        /* message buffer */
	       1,                  /* one data item */  
	       MPI_INT,            /* data item is an integer */  
	       srcProcess,         /* destination process rank */
	       WORKTAG,            /* user chosen message tag */
	       MPI_COMM_WORLD,     /* always use this */
	       &status);           /* info about received */
      
      MPI_Recv(&sizeYBlock,        /* message buffer */
	       1,                  /* one data item */  
	       MPI_INT,            /* data item is an integer */  
	       srcProcess,         /* destination process rank */
	       WORKTAG,            /* user chosen message tag */
	       MPI_COMM_WORLD,     /* always use this */
	       &status);           /* info about received */
      
      oneDimXel = (xel*) malloc(sizeXBlock*sizeYBlock*sizeof(xel));
      
      MPI_Recv(oneDimXel,          /* message buffer */
	       sizeXBlock*sizeYBlock*3, /* size of one dimension xel array */  
	       MPI_CHAR,            /* data item is character */  
	       srcProcess,         /* destination process rank */
	       WORKTAG,            /* user chosen message tag */
	       MPI_COMM_WORLD,     /* always use this */
	       &status);           /* info about received */

      // get 2D xel
      for(k=offsetX;k<(offsetX+sizeXBlock);k++)
	for(z=offsetY;z<(offsetY+sizeYBlock);z++)
	  _xel[z][k] = oneDimXel[(k-offsetX)*sizeYBlock+(z-offsetY)];

      offsetX += sizeXBlock;
      if(offsetX == colP) { offsetX = 0; }
      
      free(oneDimXel);
    }
    offsetY += sizeYBlock;
    if(offsetY == rowP) { offsetY = 0; }
  }

  // write to file
  tmpfilename = strtok(filename,".");
  extention = strtok(0,".");
  
  bzero(savefile, 35);
  sprintf(savefile, "%s_s.%s", tmpfilename, extention);

  fp = fopen(savefile, "w");
  pnm_writepnm(fp, _xel, colP, rowP, maxvalP, formatP, 0);
  fclose(fp);
}


/**
   function rendering()

**/ 

void rendering(xel **tmpxel, xel **srcxel, int Xsize, int Ysize, int iteration, int sizeOfBlock, int _rank, int masterrank) {

  MPI_Status status;

  int i,j;
  int k,z;
  int x;
  int positionX, positionY;
  xel *boundxel;
  xel *topBuff, *bottomBuff, *leftBuff, *rightBuff; 
  xel *sendxel;

  pixval red, green, blue;
  pixval northR, southR, westR, eastR;
  pixval northG, southG, westG, eastG;
  pixval northB, southB, westB, eastB;

  topBuff = (xel*) malloc(Xsize*sizeof(xel));
  bottomBuff = (xel*) malloc(Xsize*sizeof(xel));
  leftBuff = (xel*) malloc(Ysize*sizeof(xel));
  rightBuff = (xel*) malloc(Ysize*sizeof(xel));

  if (Xsize > Ysize) 
    boundxel = (xel*) malloc(Xsize*sizeof(xel));
  else
    boundxel = (xel*) malloc(Ysize*sizeof(xel));

  // calculate position itself
  positionX = (_rank/sizeOfBlock);
  positionY = (_rank%sizeOfBlock);

  for(x=0;x<iteration;x++) {

    //SEND BOUNDARY

    if(positionX!=0) {
      // SEND TOP TO UPPER NEIGHBOR
    
      for(j=0;j<Xsize;j++) 
	boundxel[j] = srcxel[0][j];
      
      MPI_Send(boundxel,           /* message buffer */
	       Xsize*3,            /* size of one dimension xel array */  
	       MPI_CHAR,           /* data item is an integer */  
	       (_rank-sizeOfBlock),/* destination process rank */
	       WORKTAG,            /* user chosen message tag */
	       MPI_COMM_WORLD);    /* always use this */    
    }

    if(positionX!=(sizeOfBlock-1)) {
      // SEND BOTTOM TO LOWER NEIGHBOR

      for(j=0;j<Xsize;j++) 
	boundxel[j] = srcxel[(Ysize-1)][j];

      MPI_Send(boundxel,           /* message buffer */
	       Xsize*3,            /* size of one dimension xel array */  
	       MPI_CHAR,           /* data item is an integer */  
	       (_rank+sizeOfBlock),/* destination process rank */
	       WORKTAG,            /* user chosen message tag */
	       MPI_COMM_WORLD);    /* always use this */    
    }
    
    if(positionY!=0) {
      // SEND LEFT BOUNDARY TO LEFT NEIGHBOR

      for(j=0;j<Ysize;j++) 
	boundxel[j] = srcxel[j][0];
      
      MPI_Send(boundxel,           /* message buffer */
	       Ysize*3,            /* size of one dimension xel array */  
	       MPI_CHAR,           /* data item is an integer */  
	       (_rank-1),          /* destination process rank */
	       WORKTAG,            /* user chosen message tag */
	       MPI_COMM_WORLD);    /* always use this */    
    }

    if(positionY!=(sizeOfBlock-1)) {
      // SEND RIGHT BOUNDARY TO RIGHT NEIGHBOR
      
      for(j=0;j<Ysize;j++) {
	boundxel[j] = srcxel[j][(Xsize-1)];
      }
      
      MPI_Send(boundxel,           /* message buffer */
	       Ysize*3,            /* size of one dimension xel array */  
	       MPI_CHAR,           /* data item is an integer */  
	       (_rank+1),          /* destination process rank */
	       WORKTAG,            /* user chosen message tag */
	       MPI_COMM_WORLD);    /* always use this */ 
    }

    // RECV BOUNDARY

    if(positionX!=0) {

      //RECV FROM UPPER NEIGHBOR
      MPI_Recv(topBuff,            /* message buffer */
	       Xsize*3,            /* data size */  
	       MPI_CHAR,           /* data item is character */  
	       (_rank-sizeOfBlock),/* receive from neighbor */
	       WORKTAG,            /* user chosen message tag */
	       MPI_COMM_WORLD,     /* always use this */
	       &status);           /* info about received */
    }
    else{
      for(j=0;j<Xsize;j++) 
	topBuff[j] = srcxel[0][j];
    }

    if(positionX!=(sizeOfBlock-1)) {

      //RECV FROM LOWER NEIGHBOR
      MPI_Recv(bottomBuff,         /* message buffer */
	       Xsize*3,            /* data size */  
	       MPI_CHAR,           /* data item is character */  
	       (_rank+sizeOfBlock),/* receive from neighbor */
	       WORKTAG,            /* user chosen message tag */
	       MPI_COMM_WORLD,     /* always use this */
	       &status);           /* info about received */
    }
    else {
      for(j=0;j<Xsize;j++) 
	bottomBuff[j] = srcxel[(Ysize-1)][j];
    }    

    if(positionY!=0) {

      //RECV FROM LEFT NEIGHBOR
      MPI_Recv(leftBuff,           /* message buffer */
	       Ysize*3,            /* data size */  
	       MPI_CHAR,           /* data item is character */  
	       (_rank-1),          /* receive from neighbor */
	       WORKTAG,            /* user chosen message tag */
	       MPI_COMM_WORLD,     /* always use this */
	       &status);           /* info about received */
    }
    else{
      for(j=0;j<Ysize;j++) {
	leftBuff[j] = srcxel[j][0];
      }
    }    

    if(positionY!=(sizeOfBlock-1)) {

      //RECV FROM RIGHT NEIGHBOR
      MPI_Recv(rightBuff,          /* message buffer */
	       Ysize*3,            /* data size */  
	       MPI_CHAR,           /* data item is character */  
	       (_rank+1),          /* receive from neighbor */
	       WORKTAG,            /* user chosen message tag */
	       MPI_COMM_WORLD,     /* always use this */
	       &status);           /* info about received */
    }
    else{
      for(j=0;j<Ysize;j++) 
	rightBuff[j] = srcxel[j][(Xsize-1)];
    }    

    for(z=0;z<Ysize;z++) {
      for(k=0;k<Xsize;k++) {

	// TOP ARRAY ACCESS BOUNDARY BUFFER
	if(z==0) {
	  northR = PPM_GETR(topBuff[k]);	  
	  northG = PPM_GETG(topBuff[k]);	  
	  northB = PPM_GETB(topBuff[k]);
	}
	else {
	  northR = PPM_GETR(srcxel[z-1][k]);	  
	  northG = PPM_GETG(srcxel[z-1][k]);	  
	  northB = PPM_GETB(srcxel[z-1][k]);
	}

	// BOTTOM ARRAY ACCESS BOUNDARY BUFFER
	if(z==(Ysize-1)) {
	  southR = PPM_GETR(bottomBuff[k]);	  
	  southG = PPM_GETG(bottomBuff[k]);	  
	  southB = PPM_GETB(bottomBuff[k]);
	}
	else {
	  southR = PPM_GETR(srcxel[z+1][k]);	  
	  southG = PPM_GETG(srcxel[z+1][k]);
	  southB = PPM_GETB(srcxel[z+1][k]);
	} 

	// LEFT ARRAY ACCESS BOUNDARY BUFFER
	if(k==0) {
	  westR = PPM_GETR(leftBuff[z]);	  
	  westG = PPM_GETG(leftBuff[z]);	  
	  westB = PPM_GETB(leftBuff[z]);
	}
	else {
	  westR = PPM_GETR(srcxel[z][k-1]);	  
	  westG = PPM_GETG(srcxel[z][k-1]);	  
	  westB = PPM_GETB(srcxel[z][k-1]);
	}

	// RIGHT ARRAY ACCESS BOUNDARY BUFFER
	if(k==(Xsize-1)) {
	  eastR = PPM_GETR(rightBuff[z]);	  
	  eastG = PPM_GETG(rightBuff[z]);	  
	  eastB = PPM_GETB(rightBuff[z]);
	}
	else {
	  eastR = PPM_GETR(srcxel[z][k+1]);	  
	  eastG = PPM_GETG(srcxel[z][k+1]);	  
	  eastB = PPM_GETB(srcxel[z][k+1]);
	}

	// get RGB value
	red = (northR + southR + westR + eastR + PPM_GETR(srcxel[z][k])) / 5;
	green = (northG + southG + westG + eastG + PPM_GETG(srcxel[z][k])) / 5;
	blue = (northB + southB + westB + eastB + PPM_GETB(srcxel[z][k])) / 5;

	PPM_ASSIGN(tmpxel[z][k], red, green, blue);
      }
    }

    // swap source xel with temp xel to iterate rendering
    for ( i = 0 ; i < Xsize ; i++ ) {
      for ( j = 0 ; j < Ysize ; j++ ) {
	srcxel[j][i] = tmpxel[j][i];
      }
    }
  }
  

  // send rendered image to master
  sendxel = (xel*) malloc(Xsize*Ysize*sizeof(xel));
  
  for(k=0;k<Xsize;k++)
    for(z=0;z<Ysize;z++)
      sendxel[(k*Ysize+z)] = srcxel[z][k];
  
  MPI_Send(&Xsize,             /* message buffer */
	   1,                  /* one data item */  
	   MPI_INT,            /* data item is an integer */  
	   masterrank,         /* destination process rank */
	   WORKTAG,            /* user chosen message tag */
	   MPI_COMM_WORLD);    /* always use this */    
  
  MPI_Send(&Ysize,             /* message buffer */
	   1,                  /* one data item */  
	   MPI_INT,            /* data item is an integer */  
	   masterrank,         /* destination process rank */
	   WORKTAG,            /* user chosen message tag */
	   MPI_COMM_WORLD);    /* always use this */    
  
  MPI_Send(sendxel,            /* message buffer */
	   Xsize*Ysize*3,      /* size of one dimension xel array */  
	   MPI_CHAR,           /* data item is character */  
	   masterrank,         /* destination process rank */
	   WORKTAG,            /* user chosen message tag */
	   MPI_COMM_WORLD);    /* always use this */    
  
  free(boundxel);
  free(sendxel);
  free(topBuff);
  free(bottomBuff);
  free(leftBuff);
  free(rightBuff);
}


/**
   function partitioning()

**/ 

void partitioning(xel **_xel, int numOfSide, int colP, int rowP, int iterations) {

  int offsetX, offsetY;
  int i, j, k, z;
  int destProcess;
  int sizeXBlock, sizeYBlock;
  int extraX, extraY;

  xel *oneDimXel;
  
  offsetX = 0;
  offsetY = 0;

  // get the margin for outline block
  extraX = colP % numOfSide;
  extraY = rowP % numOfSide;

  // calculate destination process
  for(i=0;i<numOfSide;i++) {
    for(j=0;j<numOfSide;j++) {
      destProcess = i*numOfSide + j;

      if(j==(numOfSide-1)) 
	sizeXBlock = (colP/numOfSide + extraX);
      else
	sizeXBlock = colP/numOfSide;

      if(i==(numOfSide-1))
	sizeYBlock = (rowP/numOfSide + extraY); 
      else
	sizeYBlock = rowP/numOfSide;

      oneDimXel = (xel *)malloc((sizeXBlock)*(sizeYBlock)*(sizeof(xel)));

      for(k=offsetX;k<(offsetX+sizeXBlock);k++)
	for(z=offsetY;z<(offsetY+sizeYBlock);z++)
	  oneDimXel[(k-offsetX)*sizeYBlock+(z-offsetY)]=_xel[z][k];

      MPI_Send(&sizeXBlock,        /* message buffer */
               1,                  /* one data item */  
               MPI_INT,            /* data item is an integer */  
               destProcess,        /* destination process rank */
               WORKTAG,            /* user chosen message tag */
               MPI_COMM_WORLD);    /* always use this */

      MPI_Send(&sizeYBlock,        /* message buffer */
               1,                  /* one data item */  
               MPI_INT,            /* data item is an integer */  
               destProcess,        /* destination process rank */
               WORKTAG,            /* user chosen message tag */
               MPI_COMM_WORLD);    /* always use this */

      MPI_Send(&iterations,        /* message buffer */
               1,                  /* one data item */  
               MPI_INT,            /* data item is an integer */  
               destProcess,        /* destination process rank */
               WORKTAG,            /* user chosen message tag */
               MPI_COMM_WORLD);    /* always use this */    

      MPI_Send(oneDimXel,          /* message buffer */
               sizeXBlock*sizeYBlock*3, /* size of one dimension xel array */  
               MPI_CHAR,           /* data item is character */  
               destProcess,        /* destination process rank */
               WORKTAG,            /* user chosen message tag */
               MPI_COMM_WORLD);    /* always use this */    

      offsetX += sizeXBlock;
      if(offsetX == colP) { offsetX = 0; }

      free(oneDimXel);
    }
    offsetY += sizeYBlock;
    if(offsetY == rowP) { offsetY = 0; }
  }
}


/**
   function asssembling()

**/ 

main(int argc, char **argv) {

  FILE *fp;
  char filename[30];
  int iterations;
  int rank, size;
  int colP, rowP;
  int numOfSide, sizeXBlock, sizeYBlock;
  MPI_Status status;

  xel** _xel;
  int formatP;
  xelval maxvalP;

  pnm_init( &argc, argv );
  
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);

  numOfSide = sqrt(size-1);

  // master
  if(rank == (size-1)) {  

    // get standard input
    printf("Number of Processor : %d\nNumber of Block for Each Side : %d\n",
	   size, numOfSide);
    printf("pnm File Name : ");
    scanf("%s", filename);
    printf("\n # of iterations : ");
    scanf("%d", &iterations);   
    
    // pnm file open
    fp = fopen(filename, "r");
    _xel = pnm_readpnm(fp, &colP, &rowP, &maxvalP, &formatP);
    fclose(fp);
    
    partitioning(_xel, numOfSide, colP, rowP, iterations);
    assembling(_xel, numOfSide, colP, rowP, formatP, maxvalP, filename);        
    pnm_freearray(_xel, colP);

    printf("Smoothing inputfile %s : %d times\n", filename, iterations);
  }

  // slave
  else { 

    int i, j;
    int Xsize, Ysize, iteration;
    int masterRank;
    int sizeofBlock;
    xel *oneDimXel;
    xel **srcxel, **tmpxel; 

    masterRank = (size -1);

    // get xel and parameters from function partition
    MPI_Recv(&Xsize,             /* message buffer */
	     1,                  /* one data item */  
	     MPI_INT,            /* data item is an integer */  
	     masterRank,         /* receive from master process */
	     WORKTAG,            /* user chosen message tag */
	     MPI_COMM_WORLD,     /* always use this */
	     &status);           /* info about received */

    MPI_Recv(&Ysize,             /* message buffer */
	     1,                  /* one data item */  
	     MPI_INT,            /* data item is an integer */  
	     masterRank,         /* receive from master process */
	     WORKTAG,            /* user chosen message tag */
	     MPI_COMM_WORLD,     /* always use this */
	     &status);           /* info about received */

    MPI_Recv(&iteration,         /* message buffer */
	     1,                  /* one data item */  
	     MPI_INT,            /* data item is an integer */  
	     masterRank,         /* receive from master process */
	     WORKTAG,            /* user chosen message tag */
	     MPI_COMM_WORLD,     /* always use this */
	     &status);           /* info about received */

    oneDimXel = (xel*)malloc(Xsize*Ysize*sizeof(xel));

    MPI_Recv(oneDimXel,          /* message buffer */
	     Xsize*Ysize*3,      /* size of one dimension xel array */  
	     MPI_CHAR,           /* data item is character */  
	     masterRank,         /* receive from master process */
	     WORKTAG,            /* user chosen message tag */
	     MPI_COMM_WORLD,     /* always use this */
	     &status);           /* info about received */
    
    tmpxel = (xel **) malloc(Ysize * sizeof(xel *));
    for ( j = 0 ; j < Ysize ; j++ ) {
      tmpxel[j] = (xel *) malloc(Xsize * sizeof(xel));
    }

    srcxel = (xel **) malloc(Ysize * sizeof(xel *));
    for ( j = 0 ; j < Ysize ; j++ ) {
      srcxel[j] = (xel *) malloc(Xsize * sizeof(xel));
    }

    for ( i = 0 ; i < Xsize ; i++ ) {
      for ( j = 0 ; j < Ysize ; j++ ) {
	srcxel[j][i] = oneDimXel[i*Ysize+j];
      }
    }

    rendering(tmpxel, srcxel, Xsize, Ysize, iteration, numOfSide, rank, masterRank);
  }

  MPI_Finalize();
}


